import datetime
from urllib.parse import urljoin

import validators
import requests
import html2text
import frontmatter
from flask import flash
from bs4 import BeautifulSoup
from flask import current_app

from archivy import extensions
from archivy.search import add_to_index
from archivy.data import create


class DataObj:
    __searchable__ = ["title", "content", "desc", "tags"]

    def process_bookmark_url(self):
        try:
            url_request = requests.get(self.url).text
        except Exception:
            flash(f"Could not retrieve {self.url}\n")
            self.wipe()
            return
        try:
            parsed_html = BeautifulSoup(url_request, features="html.parser")
        except Exception:
            flash(f"Could not parse {self.url}\n")
            self.wipe()
            return

        try:
            self.content = self.extract_content(parsed_html)
        except Exception:
            flash(f"Could not extract content from {self.url}\n")
            return

        parsed_title = parsed_html.title
        self.title = (parsed_title.string if parsed_title is not None
                      else self.url)

    def wipe(self):
        self.title = None
        self.desc = None
        self.content = None

    def extract_content(self, beautsoup):
        stripped_tags = ["footer", "nav"]
        url = self.url.rstrip("/")

        for tag in stripped_tags:
            if getattr(beautsoup, tag):
                getattr(beautsoup, tag).extract()
        resources = beautsoup.find_all(["a", "img"])
        for external in resources:
            if external.name == "a" and \
                    external.has_attr("href") and \
                    external["href"].startswith("/"):
                external["href"] = urljoin(url, external["href"])
            elif external.name == "img" and \
                    external.has_attr("src") and \
                    external["src"].startswith("/"):
                external["src"] = urljoin(url, external["src"])

        return html2text.html2text(str(beautsoup))

    def __init__(self, **kwargs):

        # data has already been processed
        if kwargs["type"] == "processed-dataobj":
            for key, value in kwargs.items():
                setattr(self, key, value)
        else:
            # still needs processing

            if "path" not in kwargs or kwargs["path"] == "not classified":
                kwargs["path"] = ""

            # set_attributes (path, desc, tags, type, title)
            self.path = kwargs["path"]
            self.desc = kwargs["desc"]
            self.tags = kwargs["tags"]
            self.type = kwargs["type"]
            self.title = kwargs.get("title")

            self.date = datetime.datetime.now()
            self.content = ""
            self.fullpath = ""
            self.id = None
            if self.type == "bookmarks" or self.type == "pocket_bookmarks":
                self.url = kwargs["url"]
                if validators.url(self.url):
                    self.process_bookmark_url()

    def validate(self):
        valid_url = (
            self.type != "bookmarks" or self.type != "pocket_bookmarks") or (
            isinstance(
                self.url,
                str) and validators.url(
                self.url))
        valid_title = isinstance(self.title, str)
        valid_content = (self.type != "bookmark" and
                         self.type != "pocket_bookmarks") or \
            isinstance(self.content, str)
        return valid_url and valid_title and valid_content

    def insert(self):
        if self.validate():
            extensions.set_max_id(extensions.get_max_id() + 1)
            self.id = extensions.get_max_id()
            data = {
                "type": self.type,
                "desc": self.desc,
                "title": str(self.title),
                "date": self.date.strftime("%x").replace("/", "-"),
                "tags": self.tags,
                "id": self.id,
                "path": self.path
            }
            if self.type == "bookmarks" or self.type == "pocket_bookmarks":
                data["url"] = self.url

            # convert to markdown
            dataobj = frontmatter.Post(self.content)
            dataobj.metadata = data
            self.fullpath = create(
                                frontmatter.dumps(dataobj),
                                str(self.id) + "-" +
                                dataobj["date"] + "-" + dataobj["title"],
                                path=self.path,
                                needs_to_open=self.type == "note")

            add_to_index(current_app.config['INDEX_NAME'], self)
            return self.id
        return False

    @classmethod
    def from_file(cls, filename):
        data = frontmatter.load(filename)
        dataobj = {}
        dataobj["content"] = data.content
        for pair in ["tags", "desc", "id", "title", "path"]:
            dataobj[pair] = data[pair]

        dataobj["type"] = "processed-dataobj"
        return cls(**dataobj)



import logging
from pathlib import Path
from threading import Thread

import elasticsearch
from flask import Flask

from archivy import extensions
from archivy.check_changes import run_watcher
from archivy.config import Config

app = Flask(__name__)
app.config.from_object(Config)
app.logger.setLevel(logging.INFO)

# create dir that will hold data if it doesn't already exist
DIRNAME = app.config["APP_PATH"] + "/data/"
Path(DIRNAME).mkdir(parents=True, exist_ok=True)

if app.config["ELASTICSEARCH_ENABLED"]:
    with app.app_context():
        es = extensions.get_elastic_client()

        try:
            print(
                es.indices.create(
                    index=app.config["INDEX_NAME"],
                    body=app.config["ELASTIC_CONF"]))
        except elasticsearch.ElasticsearchException:
            app.logger.info("Elasticsearch index already created")

        Thread(target=run_watcher, args=[app]).start()


app.jinja_options["extensions"].append("jinja2.ext.do")

from archivy import routes  # noqa:



from archivy.extensions import get_elastic_client


def add_to_index(index, model):
    es = get_elastic_client()
    if not es:
        return
    payload = {}
    for field in model.__searchable__:
        payload[field] = getattr(model, field)
    es.index(index=index, id=model.id, body=payload)


def remove_from_index(index, dataobj_id):
    es = get_elastic_client()
    if not es:
        return
    es.delete(index=index, id=dataobj_id)


def query_index(index, query):
    es = get_elastic_client()
    if not es:
        return []
    search = es.search(
        index=index,
        body={
            "query": {
                "multi_match": {
                    "query": query,
                    "fields": ["*"],
                    "analyzer": "rebuilt_standard"}}},
    )

    return search["hits"]["hits"]



import os
import time

import flask
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

from archivy import search, models


class ModifHandler(FileSystemEventHandler):
    def __init__(self, app: flask.Flask):
        self.app = app

    def on_modified(self, event):
        with self.app.app_context():
            if event.src_path.endswith(
                    ".md"):
                self.app.logger.info(f"Detected changes to {event.src_path}")
                dataobj = models.DataObj.from_file(event.src_path)
                search.add_to_index(self.app.config['INDEX_NAME'], dataobj)
                self.last_modify_event = event.src_path

    def on_deleted(self, event):
        with self.app.app_context():
            if event.src_path.endswith(".md"):
                id = event.src_path.split("/")[-1].split("-")[0]
                search.remove_from_index(self.app.config['INDEX_NAME'], id)
                self.app.logger.info(f"{event.src_path} has been removed")
                self.last_delete_event = event.src_path


def run_watcher(app):
    event_handler = ModifHandler(app)
    observer = Observer()
    observer.schedule(
        event_handler,
        path=os.path.join(
            app.config['APP_PATH'],
            'data/'
        ),
        recursive=True)
    observer.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()



import os
import glob
import platform
import subprocess
from pathlib import Path
from shutil import rmtree

import frontmatter
from flask import current_app
from werkzeug.utils import secure_filename


# FIXME: ugly hack to make sure the app path is evaluated at the right time
def get_data_dir():
    return os.path.join(current_app.config['APP_PATH'], "data/")

# struct to create tree like file-structure


class Directory:
    def __init__(self, name):
        self.name = name
        self.child_files = []
        self.child_dirs = {}

# method from django to sanitize filename


def get_items(collections=[], path="", structured=True):
    datacont = Directory("root") if structured else []
    if structured:
        for filename in glob.glob(get_data_dir() + path + "**/*",
                                  recursive=True):
            paths = filename.split("/data/")[1].split("/")
            data = frontmatter.load(
                filename) if filename.endswith(".md") else None

            current_dir = datacont

            # iterate through paths
            for segment in paths:
                if segment.endswith(".md"):
                    current_dir.child_files.append(data)
                else:
                    # directory has not been saved in tree yet
                    if segment not in current_dir.child_dirs:
                        current_dir.child_dirs[segment] = Directory(segment)
                    current_dir = current_dir.child_dirs[segment]
    else:
        for filename in glob.glob(get_data_dir() + path + "**/*.md",
                                  recursive=True):
            data = frontmatter.load(filename)
            if len(collections) == 0 or \
                    any([collection == data["type"]
                        for collection in collections]):
                datacont.append(data)

    return datacont


def create(contents, title, path="", needs_to_open=False):
    path_to_md_file = os.path.join(
        get_data_dir(), path, "{}.md".format(secure_filename(title)))
    with open(path_to_md_file, "w") as file:
        file.write(contents)

    if needs_to_open:
        open_file(path_to_md_file)
    return path_to_md_file


def get_item(dataobj_id):
    file = glob.glob(f"{get_data_dir()}**/{dataobj_id}-*.md",
                     recursive=True)[0]
    data = frontmatter.load(file)
    data["fullpath"] = file
    return data


def delete_item(id):
    file = glob.glob(f"{get_data_dir()}**/{id}-*.md", recursive=True)[0]
    os.remove(file)


def get_dirs():
    dirnames = glob.glob(get_data_dir() + "**/*", recursive=True)
    dirnames = [name.split("/data/")[1]
                for name in dirnames if not name.endswith(".md")]
    dirnames.append("not classified")
    return dirnames


def create_dir(name):
    sanitized_name = "/".join([secure_filename(pathname)
                               for pathname in name])
    Path(get_data_dir() + sanitized_name).mkdir(parents=True)
    print(sanitized_name)
    return sanitized_name


def delete_dir(name):
    try:
        rmtree(get_data_dir() + name)
        return True
    except FileNotFoundError:
        return False


def open_file(path):
    if platform.system() == "Windows":
        os.startfile(path)
    elif platform.system() == "Darwin":
        subprocess.Popen(["open", path])
    else:
        subprocess.Popen(["xdg-open", path])



import re
from flask_wtf import FlaskForm
from wtforms import StringField, SubmitField, SelectField
from wtforms.validators import DataRequired, URL, ValidationError


class NewBookmarkForm(FlaskForm):
    url = StringField("url", validators=[DataRequired(), URL()])
    path = SelectField("Topic")
    desc = StringField("desc")
    tags = StringField("tags")
    submit = SubmitField("Save")


class NewNoteForm(FlaskForm):
    title = StringField("title", validators=[DataRequired()])
    path = SelectField("Topic")
    desc = StringField("desc")
    tags = StringField("tags")
    submit = SubmitField("Save")


class PocketForm(FlaskForm):
    api_key = StringField("Pocket API key")
    submit = SubmitField("Save")

    def validate_api_key(self, api_key):
        key_regex = r"\d{5}-\w{24}"
        if not re.match(key_regex, api_key.data):
            raise ValidationError("Invalid API key.")


class DeleteDataForm(FlaskForm):
    submit = SubmitField("Delete")
